
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2022 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#include "pyaStatusChangedListener.h"
#include "pyaObject.h"

namespace pya
{
// --------------------------------------------------------------------------
//  Implementation of StatusChangedListener

StatusChangedListener::StatusChangedListener (PYAObjectBase *pya_object)
  : mp_pya_object (pya_object)
{
  //  .. nothing yet ..
}

void
StatusChangedListener::object_status_changed (gsi::ObjectBase::StatusEventType type)
{
  if (type == gsi::ObjectBase::ObjectDestroyed) {
    mp_pya_object->object_destroyed ();
  } else if (type == gsi::ObjectBase::ObjectKeep) {
    mp_pya_object->keep_internal ();
  } else if (type == gsi::ObjectBase::ObjectRelease) {
    mp_pya_object->release ();
  }
}

}
